
##############################################################################
# This python code includes the network architectures for the AttentionBoost
# model and the base model. It also includes parameter settings
# used in these architectures.
##############################################################################


from keras import backend as K
from keras.models import Model
from keras.layers import Input, Convolution2D, MaxPooling2D, UpSampling2D, Dropout, concatenate
from keras.layers.core import  Lambda
from keras import optimizers
from keras import objectives
from keras import metrics
import numpy as np

###############################################################################################
# y_true          : The tensor of the target values
# y_pred      		: The tensor of the corresponding predictions
# c    		  		: The tensor of the loss contribution weights of each 'y_pred' element
# w           		: A coefficient to adjust the overall loss
###############################################################################################
def weighted_mse_loss(y_true, y_pred, c, w = 1): 
    mse_loss = objectives.mean_squared_error(y_true, y_pred)
    loss = (w * c * mse_loss)
    return K.mean(loss)


###############################################################################################
# input_height       		: The height of the input
# input_width      			: The width of the input
# 
# This currying function returns next_C function which calculates next C map 
# given target map, prediction map and previous stage's C map. 
###############################################################################################
def init_next_C(input_height, input_width): 
	###########################################################################################
	# args       		: The tensors for 'y_true', 'y_pred' and 'c' values] 
	###########################################################################################
    def next_C(args):
        [y_true, y_pred, c] = args
        c = K.reshape(c, (input_height, input_width))               
        y = K.abs((y_pred-0.5))
        y = K.reshape(y, K.int_shape(c))
        
        beta_inc = (1+y) 
        beta_corr = (1-y)
        
        correct_mask = K.equal(K.round(y_pred), y_true)  # Correctly classified ones
        correct_mask = K.reshape(K.cast(correct_mask, dtype='float32'), K.int_shape(c))
    
        incorrect_mask = K.not_equal(K.round(y_pred), y_true) # incorrectly classified ones
        incorrect_mask = K.reshape(K.cast(incorrect_mask, dtype='float32'), K.int_shape(c))
        
        c_n = (c * incorrect_mask * beta_inc) + (c * correct_mask * beta_corr)    
        c_n = K.reshape(c_n, K.int_shape(c)) 
           
        #Normalize masks
        c_n = c_n * ( (incorrect_mask / K.sum(c_n * incorrect_mask)) + (correct_mask / K.sum(c_n * correct_mask)) )
        # Normalize
        c_n = (input_height * input_width * c_n) / K.sum(c_n) 
        
   
        return c_n
    
    return next_C


###############################################################################################
# This function constructs and returns a multi stage model.
###############################################################################################
# n_iterations       		: Number of iterations
# input_height       		: The height of the input
# input_width      		   : The width of the input
###############################################################################################
def attentionBoostModel( n_iterations = 4, input_height=480, input_width=640):
   
    optimizer = optimizers.Adadelta()
    
    #Inputs :
    img_gs = Input(shape= (input_height, input_width, 3))
    seg_gs = Input(shape= (input_height, input_width, 1))
    seg_null = Input(shape= (input_height, input_width, 1))
 
    # Lambda Function Init
    calc_next_c = init_next_C(input_height, input_width )

    # Models :    
    iterations=[{} for i in range(n_iterations)]
    
    for i in range(n_iterations):

            iterations[i]['toSeg'] = base_model(input_height=input_height, input_width=input_width)
            
            iterations[i]['toSeg'].name="seg_model-" + str(i)
                
    
    segmentation_outputs = []
    c = []
    shape = [input_height, input_width]

    # First C map - to balance the classes (training)
    fg_mask = K.equal(seg_gs, 1)  # foreground mask
    fg_mask = K.reshape(K.cast(fg_mask, dtype='float32'), shape)

    bg_mask = K.equal(seg_gs, 0) # background mask
    bg_mask = K.reshape(K.cast(bg_mask, dtype='float32'), shape)

    c_first = ( (fg_mask / K.sum(fg_mask)) + (bg_mask / K.sum(bg_mask)) )
    c_first = (input_height * input_width * c_first) / K.sum(c_first) # Normalize the sum
    c_first = K.reshape(c_first, shape)    
    
    c.append(c_first)
    
    for i in range(n_iterations):
        if(i == 0):
            segmentation_outputs.append(iterations[i]['toSeg']([img_gs, seg_null]))
        else:
            segmentation_outputs.append(iterations[i]['toSeg']([img_gs, segmentation_outputs[i-1]]))
            
        c.append(Lambda(calc_next_c, output_shape=(input_height,input_width))([seg_gs, segmentation_outputs[i], c[i]]) )    

       
    main_model = Model(inputs = [img_gs, seg_gs, seg_null], outputs = segmentation_outputs)
    
    for i in range(n_iterations):
        main_model.add_loss( weighted_mse_loss(seg_gs, segmentation_outputs[i], c[i], w = 1) )

    
    main_model.compile(loss=None,                        
                       optimizer= optimizer )

    return main_model





###############################################################################################
#This function constructs and returns the base FCN model used in the stages of the attentionboost_model.
#Base FCN Model: outputs segmentation map for given two inputs (image and previous segmentation map)
###############################################################################################
# input_height       	 : The height of the input
# input_width      		 : The width of the input
# k                      : The parameter to adjust the number of filters
# dropout_rate           : The dropout rate parameter
###############################################################################################

def base_model(input_height=480, input_width=640, k=32, dropout_rate = 0.2):
   
    img_input = Input(shape= (input_height, input_width, 3))
    seg_input = Input(shape= (input_height, input_width, 1))
    
    inputs = concatenate([img_input, seg_input], axis=3)
    
    conv1 = Convolution2D(k,(3, 3), activation='relu', padding='same')(inputs)
    conv1 = Dropout(dropout_rate)(conv1)
    conv1 = Convolution2D(k, (3, 3), activation='relu', padding='same')(conv1)
    pool1 = MaxPooling2D(pool_size=(2, 2))(conv1)

    conv2 = Convolution2D(k*2,(3, 3), activation='relu', padding='same')(pool1)
    conv2 = Dropout(dropout_rate)(conv2)
    conv2 = Convolution2D(k*2,(3, 3), activation='relu', padding='same')(conv2)
    pool2 = MaxPooling2D(pool_size=(2, 2))(conv2)
    
    conv3 = Convolution2D(k*4,(3, 3), activation='relu', padding='same')(pool2)
    conv3 = Dropout(dropout_rate)(conv3)
    conv3 = Convolution2D(k*4,(3, 3), activation='relu', padding='same')(conv3)
    pool3 = MaxPooling2D(pool_size=(2, 2))(conv3)
    
    conv4 = Convolution2D(k*8,(3, 3), activation='relu', padding='same')(pool3)
    conv4 = Dropout(dropout_rate)(conv4)
    conv4 = Convolution2D(k*8,(3, 3), activation='relu', padding='same')(conv4)
    pool4 = MaxPooling2D(pool_size=(2, 2))(conv4)
    
    conv5 = Convolution2D(k*16,(3, 3), activation='relu', padding='same')(pool4) #1024
    conv5 = Dropout(dropout_rate)(conv5)
    conv5 = Convolution2D(k*16,(3, 3), activation='relu', padding='same')(conv5) #1024

    up1 = concatenate([UpSampling2D(size=(2, 2))(conv5), conv4], axis=3)
    conv6 = Convolution2D(k*8,(3, 3), activation='relu', padding='same')(up1)
    conv6 = Dropout(dropout_rate)(conv6)
    conv6 = Convolution2D(k*8, (3, 3), activation='relu', padding='same')(conv6)
    
    up2 = concatenate([UpSampling2D(size=(2, 2))(conv6), conv3], axis=3)
    conv7 = Convolution2D(k*4,(3, 3), activation='relu', padding='same')(up2)
    conv7 = Dropout(dropout_rate)(conv7)
    conv7 = Convolution2D(k*4, (3, 3), activation='relu', padding='same')(conv7)
    
    up3 = concatenate([UpSampling2D(size=(2, 2))(conv7), conv2], axis=3)
    conv8 = Convolution2D(k*2,(3, 3), activation='relu', padding='same')(up3)
    conv8 = Dropout(dropout_rate)(conv8)
    conv8 = Convolution2D(k*2, (3, 3), activation='relu', padding='same')(conv8)
    
    up4 = concatenate([UpSampling2D(size=(2, 2))(conv8), conv1], axis=3)
    conv9 = Convolution2D(k,(3, 3), activation='relu', padding='same')(up4)
    conv9 = Dropout(dropout_rate)(conv9)
    conv9 = Convolution2D(k, (3, 3), activation='relu', padding='same')(conv9)

    seg_output = Convolution2D(1, (1,1), activation = 'sigmoid')(conv9)

    model = Model(inputs = [img_input, seg_input], outputs = seg_output )

    return model
